from typing import Optional, List, Tuple, IO

import os
import time
import functools
import math
from io import BytesIO
from enum import Enum
from pathlib import Path
from threading import Semaphore
from concurrent.futures import ThreadPoolExecutor, as_completed

from alipcs_py.alipcs.errors import AliPCSError
from alipcs_py.alipcs import AliPCSApi, FromTo
from alipcs_py.alipcs.pcs import CheckNameMode
from alipcs_py.common import constant
from alipcs_py.common.path import (
    is_file,
    exists,
    posix_path_basename,
    posix_path_dirname,
    walk,
)
from alipcs_py.common.event import KeyHandler, KeyboardMonitor
from alipcs_py.common.constant import CPU_NUM
from alipcs_py.common.concurrent import sure_release, retry
from alipcs_py.common.progress_bar import (
    _progress,
    progress_task_exists,
    remove_progress_task,
    reset_progress_task,
)
from alipcs_py.common.crypto import calc_sha1, calc_proof_code
from alipcs_py.common.io import (
    total_len,
    EncryptType,
    reset_encrypt_io,
)
from alipcs_py.commands.log import get_logger

from requests_toolbelt import MultipartEncoderMonitor

from rich.progress import TaskID
from rich.table import Table
from rich.box import SIMPLE
from rich.text import Text
from rich import print

logger = get_logger(__name__)

# If slice size >= 100M, the rate of uploading will be much lower.
DEFAULT_SLICE_SIZE = 80 * constant.OneM

UPLOAD_STOP = False

_rapiduploadinfo_file: Optional[str] = None


def _wait_start():
    while True:
        if UPLOAD_STOP:
            time.sleep(1)
        else:
            break


def _toggle_stop(*args, **kwargs):
    global UPLOAD_STOP
    UPLOAD_STOP = not UPLOAD_STOP
    if UPLOAD_STOP:
        print("[i yellow]Uploading stop[/i yellow]")
    else:
        print("[i yellow]Uploading continue[/i yellow]")


# Pass "p" to toggle uploading start/stop
KeyboardMonitor.register(KeyHandler("p", callback=_toggle_stop))


def adjust_slice_size(slice_size: int, io_len: int) -> int:
    """Adjust slice_size to not let part_number of the file to be larger than the 10_000"""

    part_number = math.ceil(io_len / slice_size)
    if part_number > 10_000:
        return math.ceil(io_len / 10_000)
    else:
        return slice_size


def to_remotepath(sub_path: str, remotedir: str) -> str:
    return (Path(remotedir) / sub_path).as_posix()


def from_tos(localpaths: List[str], remotedir: str) -> List[FromTo]:
    """Find all localpaths and their corresponded remotepath"""

    ft: List[FromTo] = []
    for localpath in localpaths:
        if not exists(localpath):
            continue

        if is_file(localpath):
            remotepath = to_remotepath(os.path.basename(localpath), remotedir)
            ft.append(FromTo(localpath, remotepath))
        else:
            parents_num = max(len(Path(localpath).parts) - 1, 0)
            for sub_path in walk(localpath):
                relative_path = Path(*Path(sub_path).parts[parents_num:]).as_posix()
                remotepath = to_remotepath(relative_path, remotedir)
                ft.append(FromTo(sub_path, remotepath))
    return ft


class UploadType(Enum):
    """Upload Type

    One: Upload the slices of one file concurrently
    Many: Upload files concurrently
    """

    One = 1
    Many = 2


# remotedir must be a directory
def upload(
    api: AliPCSApi,
    from_to_list: List[FromTo],
    upload_type: UploadType = UploadType.One,
    check_name_mode: CheckNameMode = "auto_rename",
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    max_workers: int = CPU_NUM,
    slice_size: int = DEFAULT_SLICE_SIZE,
    show_progress: bool = True,
    rapiduploadinfo_file: Optional[str] = None,
    user_id: Optional[str] = None,
    user_name: Optional[str] = None,
):
    """Upload from_tos

    Args:
        upload_type (UploadType): the way of uploading.
        max_workers (int): The number of concurrent workers.
        slice_size (int): The size of slice for uploading slices.
        ignore_existing (bool): Ignoring these localpath which of remotepath exist.
        show_progress (bool): Show uploading progress.
        check_name_mode(str):
          'overwrite' (直接覆盖，以后多版本有用)
          'auto_rename' (自动换一个随机名称)
          'refuse' (不会创建，告诉你已经存在)
          'ignore' (会创建重名的)
    """

    logger.debug(
        "======== Uploading start ========\n-> UploadType: %s\n-> Size of from_to_list: %s",
        upload_type,
        len(from_to_list),
    )

    global _rapiduploadinfo_file
    if _rapiduploadinfo_file is None:
        _rapiduploadinfo_file = rapiduploadinfo_file

    if upload_type == UploadType.One:
        upload_one_by_one(
            api,
            from_to_list,
            check_name_mode,
            max_workers=max_workers,
            encrypt_password=encrypt_password,
            encrypt_type=encrypt_type,
            slice_size=slice_size,
            show_progress=show_progress,
            user_id=user_id,
            user_name=user_name,
        )
    elif upload_type == UploadType.Many:
        upload_many(
            api,
            from_to_list,
            check_name_mode,
            max_workers=max_workers,
            encrypt_password=encrypt_password,
            encrypt_type=encrypt_type,
            slice_size=slice_size,
            show_progress=show_progress,
            user_id=user_id,
            user_name=user_name,
        )


def _need_to_upload(api: AliPCSApi, remotepath: str, check_name_mode: CheckNameMode) -> bool:
    """Check wether the `remotepath` needs to be uploaded

    If `check_name_mode` is `refuse` and the `remotepath` exists, then it does not need to be uploaded.
    """

    try:
        pcs_file = api.path(remotepath)
        if pcs_file and check_name_mode == "refuse":
            print(f"`{remotepath}` already exists.")
            logger.debug("`_init_encrypt_io`: remote file already exists")
            return False
        return True
    except Exception as err:
        raise err


def _init_encrypt_io(
    localpath: str,
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
) -> Tuple[IO, int, int, int]:
    assert exists(Path(localpath)), f"`{localpath}` does not exist"

    stat = Path(localpath).stat()
    local_ctime, local_mtime = int(stat.st_ctime), int(stat.st_mtime)

    encrypt_io = encrypt_type.encrypt_io(open(localpath, "rb"), encrypt_password)
    # IO Length
    encrypt_io_len = total_len(encrypt_io)

    logger.debug(
        "`_init_encrypt_io`: encrypt_type: %s, localpath: %s, encrypt_io_len: %s",
        encrypt_type,
        localpath,
        encrypt_io_len,
    )

    return (encrypt_io, encrypt_io_len, local_ctime, local_mtime)


def _rapid_upload(
    api: AliPCSApi,
    localpath: str,
    filename: str,
    dest_file_id: str,
    content_hash: str,
    proof_code: str,
    io_len: int,
    check_name_mode: CheckNameMode = "auto_rename",
    task_id: Optional[TaskID] = None,
) -> bool:
    logger.debug("`_rapid_upload`: rapid_upload starts")
    try:
        api.rapid_upload_file(
            filename,
            dest_file_id,
            io_len,
            content_hash,
            proof_code,
            check_name_mode=check_name_mode,
        )

        remove_progress_task(task_id)

        logger.debug("`_can_rapid_upload`: rapid_upload success, task_id: %s", task_id)
        return True
    except AliPCSError as err:
        logger.warning("`_can_rapid_upload`: rapid_upload fails")

        if err.error_code != 31079:  # 31079: '未找到文件MD5，请使用上传API上传整个文件。'
            remove_progress_task(task_id)
            logger.warning("`_can_rapid_upload`: unknown error: %s", err)
            raise err
        else:
            reset_progress_task(task_id)
            logger.debug("`_can_rapid_upload`: %s, no exist in remote", localpath)

        return False


def upload_one_by_one(
    api: AliPCSApi,
    from_to_list: List[FromTo],
    check_name_mode: CheckNameMode,
    max_workers: int = CPU_NUM,
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    slice_size: int = DEFAULT_SLICE_SIZE,
    show_progress: bool = True,
    user_id: Optional[str] = None,
    user_name: Optional[str] = None,
):
    """Upload files one by one with uploading the slices concurrently"""

    for from_to in from_to_list:
        task_id = None
        if show_progress:
            task_id = _progress.add_task("upload", start=False, title=from_to.from_)
        upload_file_concurrently(
            api,
            from_to,
            check_name_mode,
            max_workers=max_workers,
            encrypt_password=encrypt_password,
            encrypt_type=encrypt_type,
            slice_size=slice_size,
            task_id=task_id,
            user_id=user_id,
            user_name=user_name,
        )

    logger.debug("======== Uploading end ========")


@retry(
    -1,
    except_callback=lambda err, fail_count: logger.warning(
        "`upload_file_concurrently`: fails: error: %s, fail_count: %s",
        err,
        fail_count,
        exc_info=err,
    ),
)
def upload_file_concurrently(
    api: AliPCSApi,
    from_to: FromTo,
    check_name_mode: CheckNameMode,
    max_workers: int = CPU_NUM,
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    slice_size: int = DEFAULT_SLICE_SIZE,
    task_id: Optional[TaskID] = None,
    user_id: Optional[str] = None,
    user_name: Optional[str] = None,
):
    """Uploading one file by uploading it's slices concurrently"""

    localpath, remotepath = from_to

    remotedir = posix_path_dirname(remotepath)
    dest_dir = api.makedir_path(remotedir)
    dest_file_id = dest_dir.file_id

    filename = posix_path_basename(remotepath)

    if not _need_to_upload(api, remotepath, check_name_mode):
        remove_progress_task(task_id)
        return

    info = _init_encrypt_io(localpath, encrypt_password=encrypt_password, encrypt_type=encrypt_type)
    encrypt_io, encrypt_io_len, local_ctime, local_mtime = info
    slice_size = adjust_slice_size(slice_size, encrypt_io_len)
    part_number = math.ceil(encrypt_io_len / slice_size)

    # Progress bar
    if task_id is not None and progress_task_exists(task_id):
        _progress.update(task_id, total=encrypt_io_len)
        _progress.start_task(task_id)

    slice_completed = 0
    slice_completeds = {}  # current i-th index slice completed size

    def callback_for_slice(upload_url: str, monitor: MultipartEncoderMonitor):
        if task_id is not None and progress_task_exists(task_id):
            slice_completeds[upload_url] = monitor.bytes_read
            current_compledted: int = sum(list(slice_completeds.values()))
            _progress.update(task_id, completed=slice_completed + current_compledted)

    slice1k_hash = ""
    content_hash = ""

    pcs_prepared_file = None
    if encrypt_type == EncryptType.No and encrypt_io_len >= 1 * constant.OneK:
        # Rapid Upload

        slice1k_bytes = encrypt_io.read(constant.OneK)
        reset_encrypt_io(encrypt_io)
        slice1k_hash = calc_sha1(slice1k_bytes)

        pcs_prepared_file = api.prepare_file(
            filename,
            dest_file_id,
            encrypt_io_len,
            slice1k_hash,
            part_number=part_number,
            check_name_mode=check_name_mode,
        )
        if pcs_prepared_file.can_rapid_upload():
            content_hash = calc_sha1(encrypt_io)
            proof_code = calc_proof_code(encrypt_io, encrypt_io_len, api.access_token)

            # Rapid upload
            _rapid_upload(
                api,
                localpath,
                filename,
                dest_file_id,
                content_hash,
                proof_code,
                encrypt_io_len,
                check_name_mode=check_name_mode,
                task_id=task_id,
            )
            return

    try:
        # Upload file slice
        logger.debug("`upload_file_concurrently`: upload_slice starts")

        if not pcs_prepared_file:
            pcs_prepared_file = api.create_file(
                filename,
                dest_file_id,
                encrypt_io_len,
                part_number=part_number,
                check_name_mode=check_name_mode,
            )

        reset_encrypt_io(encrypt_io)

        def _upload_slice(item):
            if not item:
                return

            io, upload_url = item

            # Retry upload until success
            retry(
                -1,
                except_callback=lambda err, fail_count: (
                    io.seek(0, 0),
                    logger.warning(
                        "`upload_file_concurrently`: error: %s, fail_count: %s",
                        err,
                        fail_count,
                        exc_info=err,
                    ),
                    _wait_start(),
                ),
            )(api.upload_slice)(
                io,
                upload_url,
                callback=functools.partial(callback_for_slice, upload_url),
            )

            slice_completeds.pop(upload_url)

            nonlocal slice_completed
            slice_completed += total_len(io)

        semaphore = Semaphore(max_workers)
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futs = []
            i = 0
            for upload_url in pcs_prepared_file.upload_urls():
                semaphore.acquire()

                size = min(slice_size, encrypt_io_len - i)
                if size == 0:
                    break

                data = encrypt_io.read(size)
                io = BytesIO(data or b"")

                fut = executor.submit(sure_release, semaphore, _upload_slice, (io, upload_url))
                futs.append(fut)

                i += size

            as_completed(futs)

        file_id = pcs_prepared_file.file_id
        upload_id = pcs_prepared_file.upload_id
        assert file_id and upload_id
        api.upload_complete(file_id, upload_id)

        remove_progress_task(task_id)

        logger.debug(
            "`upload_file_concurrently`: upload_slice and combine_slices success, task_id: %s",
            task_id,
        )
    except Exception as err:
        logger.warning("`upload_file_concurrently`: error: %s", err)
        raise err
    finally:
        encrypt_io.close()
        reset_progress_task(task_id)


def upload_many(
    api: AliPCSApi,
    from_to_list: List[FromTo],
    check_name_mode: CheckNameMode = "overwrite",
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    max_workers: int = CPU_NUM,
    slice_size: int = DEFAULT_SLICE_SIZE,
    show_progress: bool = True,
    rapiduploadinfo_file: Optional[str] = None,
    user_id: Optional[str] = None,
    user_name: Optional[str] = None,
):
    """Upload files concurrently that one file is with one connection"""

    excepts = {}
    semaphore = Semaphore(max_workers)
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futs = {}
        for idx, from_to in enumerate(from_to_list):
            semaphore.acquire()
            task_id = None
            if show_progress:
                task_id = _progress.add_task("upload", start=False, title=from_to.from_)

            logger.debug("`upload_many`: Upload: index: %s, task_id: %s", idx, task_id)

            fut = executor.submit(
                sure_release,
                semaphore,
                upload_file,
                api,
                from_to,
                check_name_mode,
                encrypt_password=encrypt_password,
                encrypt_type=encrypt_type,
                slice_size=slice_size,
                task_id=task_id,
                user_id=user_id,
                user_name=user_name,
            )
            futs[fut] = from_to

        for fut in as_completed(futs):
            e = fut.exception()
            if e is not None:
                from_to = futs[fut]
                excepts[from_to] = e

    logger.debug("======== Uploading end ========")

    # Summary
    if excepts:
        table = Table(title="Upload Error", box=SIMPLE, show_edge=False)
        table.add_column("From", justify="left", overflow="fold")
        table.add_column("To", justify="left", overflow="fold")
        table.add_column("Error", justify="left")

        for from_to, e in sorted(excepts.items()):
            table.add_row(from_to.from_, Text(str(e), style="red"))

        _progress.console.print(table)


@retry(
    -1,
    except_callback=lambda err, fail_count: logger.warning(
        "`upload_file`: fails: error: %s, fail_count: %s",
        err,
        fail_count,
        exc_info=err,
    ),
)
def upload_file(
    api: AliPCSApi,
    from_to: FromTo,
    check_name_mode: CheckNameMode,
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    slice_size: int = DEFAULT_SLICE_SIZE,
    task_id: Optional[TaskID] = None,
    user_id: Optional[str] = None,
    user_name: Optional[str] = None,
):
    """Upload one file with one connection"""

    _wait_start()

    localpath, remotepath = from_to

    remotedir = posix_path_dirname(remotepath)
    dest_dir = api.makedir_path(remotedir)
    dest_file_id = dest_dir.file_id

    filename = posix_path_basename(remotepath)

    if not _need_to_upload(api, remotepath, check_name_mode):
        remove_progress_task(task_id)
        return

    info = _init_encrypt_io(localpath, encrypt_password=encrypt_password, encrypt_type=encrypt_type)
    encrypt_io, encrypt_io_len, local_ctime, local_mtime = info
    slice_size = adjust_slice_size(slice_size, encrypt_io_len)
    part_number = math.ceil(encrypt_io_len / slice_size)

    # Progress bar
    if task_id is not None and progress_task_exists(task_id):
        _progress.update(task_id, total=encrypt_io_len)
        _progress.start_task(task_id)

    slice_completed = 0

    def callback_for_slice(monitor: MultipartEncoderMonitor):
        if task_id is not None and progress_task_exists(task_id):
            _progress.update(task_id, completed=slice_completed + monitor.bytes_read)

    slice1k_hash = ""
    content_hash = ""

    pcs_prepared_file = None
    if encrypt_type == EncryptType.No and encrypt_io_len >= 1 * constant.OneK:
        # Rapid Upload

        slice1k_bytes = encrypt_io.read(constant.OneK)
        reset_encrypt_io(encrypt_io)
        slice1k_hash = calc_sha1(slice1k_bytes)

        pcs_prepared_file = api.prepare_file(
            filename,
            dest_file_id,
            encrypt_io_len,
            slice1k_hash,
            part_number=part_number,
            check_name_mode=check_name_mode,
        )
        if pcs_prepared_file.can_rapid_upload():
            content_hash = calc_sha1(encrypt_io)
            proof_code = calc_proof_code(encrypt_io, encrypt_io_len, api.access_token)

            # Rapid upload
            _rapid_upload(
                api,
                localpath,
                filename,
                dest_file_id,
                content_hash,
                proof_code,
                encrypt_io_len,
                check_name_mode=check_name_mode,
                task_id=task_id,
            )
            return

    try:
        # Upload file slice
        logger.debug("`upload_file`: upload_slice starts")

        if not pcs_prepared_file:
            pcs_prepared_file = api.create_file(
                filename,
                dest_file_id,
                encrypt_io_len,
                part_number=part_number,
                check_name_mode=check_name_mode,
            )

        reset_encrypt_io(encrypt_io)

        for upload_url in pcs_prepared_file.upload_urls():
            _wait_start()

            logger.debug("`upload_file`: upload_slice: slice_completed: %s", slice_completed)

            size = min(slice_size, encrypt_io_len - slice_completed)
            if size == 0:
                break

            data = encrypt_io.read(size) or b""
            io = BytesIO(data)

            logger.debug("`upload_file`: upload_slice: size should be %s == %s", size, len(data))

            # Retry upload until success
            retry(
                -1,
                except_callback=lambda err, fail_count: (
                    io.seek(0, 0),
                    logger.warning(
                        "`upload_file`: `upload_slice`: error: %s, fail_count: %s",
                        err,
                        fail_count,
                        exc_info=err,
                    ),
                    _wait_start(),
                ),
            )(api.upload_slice)(io, upload_url, callback=callback_for_slice)

            slice_completed += size

        file_id = pcs_prepared_file.file_id
        upload_id = pcs_prepared_file.upload_id
        assert file_id and upload_id
        api.upload_complete(file_id, upload_id)

        remove_progress_task(task_id)

        logger.debug(
            "`upload_file`: upload_slice and combine_slices success, task_id: %s",
            task_id,
        )
    except Exception as err:
        logger.warning("`upload_file`: error: %s", err)
        raise err
    finally:
        encrypt_io.close()
        reset_progress_task(task_id)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['alipcs_py',
 'alipcs_py.alipcs',
 'alipcs_py.app',
 'alipcs_py.commands',
 'alipcs_py.common',
 'alipcs_py.config',
 'alipcs_py.storage']

package_data = \
{'': ['*']}

install_requires = \
['aget>=0.1',
 'chardet>=5.0',
 'click>=8.1',
 'cryptography>=38.0',
 'cython>=0.29',
 'ecdsa>=0.18',
 'fastapi>=0.87',
 'jinja2>=3.1',
 'passlib>=1.7',
 'peewee>=3.15',
 'pillow>=9.3',
 'qrcode>=7.4',
 'requests-toolbelt>=0.10',
 'requests>=2.28',
 'rich>=12.6',
 'toml>=0.10',
 'typing-extensions>=4.4',
 'uvicorn>=0.19']

entry_points = \
{'console_scripts': ['AliPCS-Py = alipcs_py.app:main']}

setup_kwargs = {
    'name': 'alipcs-py',
    'version': '0.6.1',
    'description': 'Ali Pcs Api and App',
    'long_description': '# AliPCS-Py\n\n**2023-04-24 支持 阿里云盘开放平台 api。**\n\n**2023-02-15 使用临时 API 接口，让下载可用。**\n\n**2023-02-14 阿里网盘 API 下载接口不再提供第三方应用使用。需要申请使用官方 API 接口。目前官方 API 接口在内测中。本项目已提交内测申请，等待回复中。在此期间下载功能不能使用。**\n\n[![PyPI version](https://badge.fury.io/py/alipcs-py.svg)](https://badge.fury.io/py/alipcs-py)\n![Build](https://github.com/PeterDing/AliPCS-Py/workflows/AliPCS-Py%20Build%20&%20Test/badge.svg)\n\nAn AliPCS API and An App\n\nAliPCS-Py 是阿里云盘的非官方 api 和一个命令行运用程序。\n\n---\n\n## 百度云盘 api 和 命令行客户端在 https://github.com/PeterDing/BaiduPCS-Py\n\n---\n\n- [安装](#安装)\n- [更新](#更新)\n- [API](#API)\n- [用法](#用法)\n- [命令别名](#命令别名)\n- [对多个帐号进行相同操作](#对多个帐号进行相同操作)\n\n#### 用户相关命令\n\n- [添加用户](#添加用户)\n- [显示当前用户的信息](#显示当前用户的信息)\n- [更新用户信息](#更新用户信息)\n- [显示所有用户](#显示所有用户)\n- [切换当前用户](#切换当前用户)\n- [删除一个用户](#删除一个用户)\n- [显示当前工作目录](#显示当前工作目录)\n- [切换当前工作目录](#切换当前工作目录)\n\n#### 无感加密解密文件\n\n- [设置文件加密密码](#设置文件加密密码)\n\n#### 文件操作相关命令\n\n- [文件操作](#文件操作)\n- [列出网盘路径下的文件](#列出网盘路径下的文件)\n- [搜索文件](#搜索文件)\n- [显示文件内容](#显示文件内容)\n- [创建目录](#创建目录)\n- [移动文件](#移动文件)\n- [文件重命名](#文件重命名)\n- [拷贝文件](#拷贝文件)\n- [删除文件](#删除文件)\n- [下载文件或目录](#下载文件或目录)\n- [播放媒体文件](#播放媒体文件)\n- [上传文件](#上传文件)\n- [同步本地目录到远端](#同步本地目录到远端)\n\n#### 秒传\n\n- [关于秒传连接](#关于秒传连接)\n\n#### 分享相关命令\n\n- [分享文件](#分享文件)\n- [列出分享链接](#列出分享链接)\n- [取消分享链接](#取消分享链接)\n- [列出其他用户分享链接中的文件](#列出其他用户分享链接中的文件)\n- [下载他人分享的文件](#下载他人分享的文件)\n- [播放他人分享的文件](#播放他人分享的文件)\n- [保存其他用户分享的链接](#保存其他用户分享的链接)\n\n#### 本地保存分享连接\n\n- [保存分享连接至本地](#保存分享连接至本地)\n- [显示本地保存的分享连接](#显示本地保存的分享连接)\n- [显示本地保存的分享文件](#显示本地保存的分享文件)\n- [查找本地保存的分享连接](#查找本地保存的分享连接)\n- [查找本地保存的分享文件](#查找本地保存的分享文件)\n- [查找本地保存的分享连接和文件](#查找本地保存的分享连接和文件)\n- [删除本地保存的分享连接或文件](#删除本地保存的分享连接或文件)\n- [清理本地保存的无效分享连接](#清理本地保存的无效分享连接)\n\n#### HTTP 服务\n\n- [开启 HTTP 服务](#开启-HTTP-服务)\n\n## 安装\n\n需要 Python 版本大于或等于 3.8\n\n```\npip3 install Cython\npip3 install AliPCS-Py\n```\n\n### Windows 依赖\n\n在 Windows 上，AliPCS-Py 依赖 [Microsoft C++ Build Tools](https://visualstudio.microsoft.com/visual-cpp-build-tools/)。\n\n在安装 AliPCS-Py 前，请先安装 [Microsoft C++ Build Tools](https://visualstudio.microsoft.com/visual-cpp-build-tools/)，再在其中勾选 `C++ 生成工具` 并安装。完成后即可安装 AliPCS-Py。\n\n## 更新\n\n```\npip3 install AliPCS-Py --upgrade\n```\n\n## API\n\nAliPCS-Py 的阿里云盘 API 只依赖 requests，方便用户开发自己的运用。\n\n```python\nfrom alipcs_py.alipcs import AliPCSApi\n\napi = AliPCSApi(refresh_token)\n```\n\n## 用法\n\n```\nAliPCS-Py --help\n```\n\n## 命令别名\n\n可以用下面的命令别名代替原来的命令名。\n\n| 别名 | 原名         |\n| ---- | ------------ |\n| w    | who          |\n| uu   | updateuser   |\n| su   | su           |\n| ul   | userlist     |\n| ua   | useradd      |\n| ep   | encryptpwd   |\n| ud   | userdel      |\n| l    | ls           |\n| f    | search       |\n| md   | mkdir        |\n| mv   | move         |\n| rn   | rename       |\n| cp   | copy         |\n| rm   | remove       |\n| d    | download     |\n| p    | play         |\n| u    | upload       |\n| sn   | sync         |\n| S    | share        |\n| sl   | shared       |\n| cs   | cancelshared |\n| s    | save         |\n| sv   | server       |\n\n## 对多个帐号进行相同操作\n\nAliPCS-Py 支持对多个帐号进行相同操作。比如，用相同关键字搜索多个帐号，上传相同的文件/目录到多个帐号，等等。\n\n使用者只需用 `--users` 选项来指定要操作的帐号名即可。\n\n`--users` 接受一个参数，这个参数是用“,”连接的要进行操作帐号名的部分字符。假设我们现在有 3 个帐号，帐号名分别是 `Tom`，`Peter`，`Joy`。\n现在我要同时对`Tom`和`Joy`进行关键字搜索。我们可以用下面的命令进行：\n\n```\nAliPCS-Py --users \'Tom,Joy\' search \'keyword\' / -R\n```\n\n或者给出帐号名的部分片段：\n\n```\nAliPCS-Py --users \'om,oy\' search \'keyword\' / -R\n```\n\n更简单可以用：\n\n```\n# Tom, Joy 都包含字符 "o"\nAliPCS-Py --users \'o\' search \'keyword\' / -R\n```\n\n如果要对所有帐号进行操作用 `--users \'\'`。\n\n如果不使用 `--users` 选项，默认只对当前帐号进行操作。\n\n以下命令支持对多个帐号进行操作：\n\n- pwd\n- ls\n- search\n- cat\n- mkdir\n- move\n- rename\n- copy\n- remove\n- download\n- play\n- upload\n- sync\n- share\n- shared\n- cancelshared\n- save\n- server\n\n**注意**: `--users` 一定要跟在 `AliPCS-Py` 后，命令前。\n\n## 添加用户\n\n**从 2023-02-14 开始，阿里云盘官方限制了 web 端 api 的调用。从 web 端 api 获取到的下载连接是限速的。但如果调用[阿里云盘开放平台](https://survey.alibaba.com/apps/zhiliao/I9Dd1Nl89)的 api 获取到的下载连接是不限速的。**\n\nAliPCS-Py (>= v0.6.0) 支持调用阿里云盘开放平台 api。但是由于一直没有拿到内测，没法提供默认登录操作。需要用户自己找其他应用提供的登录方式登录。\n\n### 使用 web `refresh_token` 和 第三方认证地址 登录\n\n第三方认证地址提供阿里云盘开放平台的认证服务。由于一直没有拿到内测，本项目目前没法提供。需要使用者自行寻找。\n\n交互添加：\n\n```\nAliPCS-Py useradd\n```\n\n或者直接添加：\n\n```\nAliPCS-Py useradd --web-refresh-token "..." --client-server "..."\n# 其他选项留空\n```\n\n之后用阿里云盘 APP 扫码登录。\n\n### 使用 web `refresh_token` 和 阿里云盘开放平台认证凭证 登录\n\n如果使用者拿到了阿里云盘开放平台认证，会获得 `client-id` 和 `client-secret`。使用这两个值可以直接登录。\n\n交互添加：\n\n```\nAliPCS-Py useradd\n```\n\n或者直接添加：\n\n```\nAliPCS-Py useradd --web-refresh-token "..." --client-id "..." --client-secret "..."\n# 其他选项留空\n```\n\n之后用阿里云盘 APP 扫码登录。\n\n### 使用 web `refresh_token` 登录\n\n使用者需要在 https://www.aliyundrive.com/drive/ 登录后获取 `refresh_token` 值，并用命令 `useradd` 为 AliPCS-Py 添加一个用户。`useradd` 其他参数留空就好。\n\n使用者可以用下面的方式获取用户的 `refresh_token` 值。\n\n1. 登录 https://www.aliyundrive.com/drive/\n2. 打开浏览器的开发者工具(如 Chrome DevTools)。\n3. 然后选择开发者工具的 Console 面板。输入 `JSON.parse(localStorage.token).refresh_token`，再回车，获取 `refresh_token`。\n\n![refresh_token](./imgs/refresh_token.png)\n\n现在找到了 `refresh_token` 值，我们可以用下面的命令添加一个用户。\n\n交互添加：\n\n```\nAliPCS-Py useradd\n```\n\n或者直接添加：\n\n```\nAliPCS-Py useradd --web-refresh-token "..."\n```\n\nAliPCS-Py 支持多用户，你只需一直用 `useradd` 来添加用户即可。\n\n**注意：如果只用 `--web-refresh-token` 登录，下载文件时，服务器端会限速。**\n\n## 显示当前用户的信息\n\n```\nAliPCS-Py who\n```\n\n或者：\n\n```\nAliPCS-Py who user_id\n```\n\n指明显示用户 id 为 `user_id` 的用户信息。\n\n### 选项\n\n| Option                       | Description                |\n| ---------------------------- | -------------------------- |\n| -K, --show-encrypt-password  | 显示加密密码               |\n| --account-name TEXT          | 账号名 [默认为 user id]    |\n| --web-refresh-token TEXT     | 用户 web_refresh_token     |\n| --openapi-refresh-token TEXT | 用户 openapi_refresh_token |\n| --client-id TEXT             | openapi client id          |\n| --client-secret TEXT         | openapi client secret      |\n| --client-server TEXT         | openapi client server      |\n\n## 更新用户信息\n\n默认更新当前用户信息。\n\n```\nAliPCS-Py updateuser\n```\n\n也可指定多个 `user_id`\n\n```\nAliPCS-Py updateuser user_id\n```\n\n## 显示所有用户\n\n```\nAliPCS-Py userlist\n```\n\n## 切换当前用户\n\n```\nAliPCS-Py su\n```\n\n或者指定用户列表中用户所在的位置：\n\n```\nAliPCS-Py su 2\n```\n\n## 删除一个用户\n\n```\nAliPCS-Py userdel\n```\n\n## 设置文件加密密码\n\nAliPCS-Py 支持“无感的”文件加密。\n\nAliPCS-Py 可以加密上传文件，在下载的时候自动解密，让使用者感觉不到加密解密的过程。\n\n如果使用者需要将保密文件上传至阿里云盘保存，可以使用这个方法。即使帐号被盗，攻击者也无法还原文件内容。\n\nAliPCS-Py 支持以下加密方法：\n\n- **Simple** 一种简单的加密算法。根据密钥生成一个字节对照表来加密解密文件。\n  速度快，但**不安全**，不建议加密重要文件。\n  因为这种算法加解密不需要知道上下文信息，所以，下载时支持分段下载，如果是媒体文件则支持拖动播放。\n  推荐用于加密不重要的媒体文件。\n- **ChaCha20** 工业级加密算法，速度快，推荐用于加密重要文件。不支持分段下载。\n- **AES256CBC** 工业级加密算法，推荐用于加密重要文件。不支持分段下载。\n\n**注意**：用命令 `encryptpwd` 设置的密码**只是为当前用户**的。\n\n为当前用户设置加密密码:\n\n交互添加：\n\n```\nAliPCS-Py encryptpwd\n```\n\n或者直接添加：\n\n```\nAliPCS-Py encryptpwd --encrypt-password \'my-encrypt-password\'\n```\n\n上传并加密文件：\n\n上传和同步文件时只需要指定加密算法就可。如果不指定就不加密。\n\n```\n# 默认使用上面设置的 `encrypt-password`\nAliPCS-Py upload some-file.mp4 some-dir/ /to/here --encrypt-type AES256CBC\n```\n\n下载并用上面设置的 `encrypt-password` 自动解密文件：\n\n```\nAliPCS-Py download /to/here/some-file.mp4 /to/here/some-dir/\n```\n\n也可以使用临时的 `encrypt-password`：\n\n```\nAliPCS-Py upload some-file.mp4 some-dir/ /to/here --encrypt-type Simple --encrypt-password \'onlyyou\'\n```\n\n但在使用临时的 `encrypt-password` 后，`cat`、下载和播放这些文件时需要指定 `encrypt-password`，但不需要指定加密算法，程序会自动检查加密算法：\n\n```\n# 下载\nAliPCS-Py download /to/here/some-file.mp4 /to/here/some-dir/  --encrypt-password \'onlyyou\'\n\n# 开启本地服务并播放\nAliPCS-Py play /to/here/some-file.mp4 --encrypt-password \'onlyyou\' --use-local-server\n```\n\n显示当前用户的密钥：\n\n```\nAliPCS-Py who --show-encrypt-password\n```\n\nAliPCS-Py 下载时默认会解密文件，如果想要下载但不解密文件，需要加 `--no-decrypt`\n\n```\nAliPCS-Py download some-file --no-decrypt\n```\n\n## 文件操作\n\nAliPCS-Py 操作网盘中的文件可以使用文件的绝对路径或相对路径（相对与当前目录 pwd）。\n\n每一个用户都有自己的当前工作目录（pwd），默认为 `/` 根目录。\n\n使用者可以用 `cd` 命令来切换当前的工作目录（pwd）。\n\n下面所有涉及网盘路径的命令，其中如果网盘路径用的是相对路径，那么是相对于当前工作目录（pwd）的。\n如果是网盘路径用的是绝对路径，那么就是这个绝对路径。\n\n## 显示当前工作目录\n\n```\nAliPCS-Py pwd\n```\n\n## 切换当前工作目录\n\n切换到绝对路径：\n\n```\nAliPCS-Py cd /to/some/path\n```\n\n切换到相对路径：\n\n```\n# 切换到 (pwd)/../path\nAliPCS-Py cd ../path\n```\n\n## 列出网盘路径下的文件\n\n使用文件路径：\n\n```\nAliPCS-Py ls [OPTIONS] [REMOTEPATHS]...\n\nAliPCS-Py ls /absolute/path\n\n# or\nAliPCS-Py ls relative/path\n```\n\n使用文件 ID：\n\n```\nAliPCS-Py ls -i file_id1 -i file_id2 -i ...\n```\n\n### 选项\n\n| Option                     | Description                                           |\n| -------------------------- | ----------------------------------------------------- |\n| -i, --file-id TEXT         | 文件 ID                                               |\n| --share-id, --si TEXT      | 列出这个分享 ID 下的文件                              |\n| --share-url, --su TEXT     | 列出这个分享 url 下的文件                             |\n| -p, --password TEXT        | 分享链接密码，如果没有不用设置                        |\n| -r, --desc                 | 逆序排列文件                                          |\n| -n, --name                 | 依名字排序                                            |\n| -t, --time                 | 依时间排序                                            |\n| -s, --size                 | 依文件大小排序                                        |\n| -R, --recursive            | 递归列出文件                                          |\n| -I, --include TEXT         | 筛选包含这个字符串的文件                              |\n| --include-regex, --IR TEXT | 筛选包含这个正则表达式的文件                          |\n| -E, --exclude TEXT         | 筛选 **不** 包含这个字符串的文件                      |\n| --exclude-regex, --ER TEXT | 筛选 **不** 包含这个正则表达式的文件                  |\n| -f, --is-file              | 筛选 **非** 目录文件                                  |\n| -d, --is-dir               | 筛选目录文件                                          |\n| --no-highlight, --NH       | 取消匹配高亮                                          |\n| -S, --show-size            | 显示文件大小                                          |\n| -D, --show-date            | 显示文件创建时间                                      |\n| --show-file-id, --ID       | 显示文件 ID                                           |\n| -M, --show-hash            | 显示文件 sha1                                         |\n| -A, --show-absolute-path   | 显示文件绝对路径                                      |\n| --show-dl-link, --DL       | 显示文件下载连接                                      |\n| --csv                      | 用 csv 格式显示，单行显示，推荐和 --DL 或 --HL 一起用 |\n| --only-dl-link, --ODL      | 只显示文件下载连接                                    |\n\n## 搜索文件\n\n搜索包含 `keyword` 的文件\n\n```\nAliPCS-Py search [OPTIONS] KEYWORD [REMOTEDIR]\n\n# 在当前工作目录中搜索\nAliPCS-Py search keyword\n\n# or\nAliPCS-Py search keyword /absolute/path\n\n# or\nAliPCS-Py search keyword relative/path\n```\n\n### 选项\n\n| Option                     | Description                          |\n| -------------------------- | ------------------------------------ |\n| -R, --recursive            | 递归搜索文件                         |\n| -I, --include TEXT         | 筛选包含这个字符串的文件             |\n| --include-regex, --IR TEXT | 筛选包含这个正则表达式的文件         |\n| -E, --exclude TEXT         | 筛选 **不** 包含这个字符串的文件     |\n| --exclude-regex, --ER TEXT | 筛选 **不** 包含这个正则表达式的文件 |\n| -f, --is-file              | 筛选 **非** 目录文件                 |\n| -d, --is-dir               | 筛选目录文件                         |\n| --no-highlight, --NH       | 取消匹配高亮                         |\n| -S, --show-size            | 显示文件大小                         |\n| -D, --show-date            | 显示文件创建时间                     |\n| -M, --show-hash            | 显示文件 sha1                        |\n| --csv                      | 用 csv 格式显示                      |\n\n## 显示文件内容\n\n```\nAliPCS-Py cat [OPTIONS] REMOTEPATH\n```\n\n### 选项\n\n| Option                        | Description                  |\n| ----------------------------- | ---------------------------- |\n| -e, --encoding TEXT           | 文件编码，默认自动解码       |\n| --no-decrypt, --ND            | 不解密                       |\n| --encrypt-password, --ep TEXT | 加密密码，默认使用用户设置的 |\n\n## 创建目录\n\n```\nAliPCS-Py mkdir [OPTIONS] [REMOTEDIRS]...\n```\n\n### 选项\n\n| Option     | Description |\n| ---------- | ----------- |\n| -S, --show | 显示目录    |\n\n## 移动文件\n\n移动一些文件到一个目录中。\n\n```\nAliPCS-Py move [OPTIONS] [REMOTEPATHS]... REMOTEDIR\n```\n\n### 选项\n\n| Option     | Description |\n| ---------- | ----------- |\n| -S, --show | 显示结果    |\n\n## 文件重命名\n\n```\nAliPCS-Py rename [OPTIONS] REMOTEPATH NEW_NAME\n```\n\ne.g.\n\n重命名 `/path/to/far` to `/path/to/foo`\n\n```\nAliPCS-Py rename /path/to/far foo\n```\n\n### 选项\n\n| Option     | Description |\n| ---------- | ----------- |\n| -S, --show | 显示结果    |\n\n## 拷贝文件\n\n拷贝一些文件到一个目录中。\n\n```\nAliPCS-Py move [OPTIONS] [REMOTEPATHS]... REMOTEDIR\n```\n\n### 选项\n\n| Option     | Description |\n| ---------- | ----------- |\n| -S, --show | 显示结果    |\n\n## 删除文件\n\n```\nAliPCS-Py remove [OPTIONS] [REMOTEPATHS]...\n\n# 指定 路径\nAliPCS-Py remove /some/path\n# 指定 file-id\nAliPCS-Py remove --file-id ...\n```\n\n### 选项\n\n| Option        | Description  |\n| ------------- | ------------ |\n| -i, --file-id | TEXT 文件 ID |\n\n## 下载文件或目录\n\n> 在使用了阿里云盘开放平台 api 时，如果下载他人分享的文件，文件会先保存到 `/__alipcs_py_temp__`，在从其中下载。\n> 下载结束后，会删除 `__alipcs_py_temp__` 中下载的文件。\n\n使用文件路径：\n\n```\nAliPCS-Py download [OPTIONS] [REMOTEPATHS]...\n```\n\n使用文件 ID：\n\n```\nAliPCS-Py download -i file_id1 -i file_id2 -i ...\n```\n\n### 选项\n\n| Option                                                 | Description                                                                                                                                                                                                                                                                                                                                                                                                                                                    |\n| ------------------------------------------------------ | -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |\n| -i, --file-id TEXT                                     | 文件 ID                                                                                                                                                                                                                                                                                                                                                                                                                                                        |\n| --share-id, --si TEXT                                  | 下载这个分享 ID 下的文件                                                                                                                                                                                                                                                                                                                                                                                                                                       |\n| --share-url, --su TEXT                                 | 下载这个分享 url 下的文件                                                                                                                                                                                                                                                                                                                                                                                                                                      |\n| -p, --password TEXT                                    | 分享链接密码，如果没有不用设置                                                                                                                                                                                                                                                                                                                                                                                                                                 |\n| -o, --outdir TEXT                                      | 指定下载本地目录，默认为当前目录                                                                                                                                                                                                                                                                                                                                                                                                                               |\n| -R, --recursive                                        | 递归下载                                                                                                                                                                                                                                                                                                                                                                                                                                                       |\n| -f, --from-index INTEGER                               | 从所有目录中的第几个文件开始下载，默认为 0（第一个）                                                                                                                                                                                                                                                                                                                                                                                                           |\n| -I, --include TEXT                                     | 筛选包含这个字符串的文件                                                                                                                                                                                                                                                                                                                                                                                                                                       |\n| --include-regex, --IR TEXT                             | 筛选包含这个正则表达式的文件                                                                                                                                                                                                                                                                                                                                                                                                                                   |\n| -E, --exclude TEXT                                     | 筛选 不 包含这个字符串的文件                                                                                                                                                                                                                                                                                                                                                                                                                                   |\n| --exclude-regex, --ER TEXT                             | 筛选 不 包含这个正则表达式的文件                                                                                                                                                                                                                                                                                                                                                                                                                               |\n| -s, --concurrency INTEGER                              | 下载同步链接数，默认为 5。建议小于 10。                                                                                                                                                                                                                                                                                                                                                                                                                        |\n| -k, --chunk-size TEXT                                  | 同步链接分块大小，默认为 50MB                                                                                                                                                                                                                                                                                                                                                                                                                                  |\n| -q, --quiet                                            | 取消第三方下载应用输出                                                                                                                                                                                                                                                                                                                                                                                                                                         |\n| --out-cmd, --OC                                        | 输出第三方下载应用命令                                                                                                                                                                                                                                                                                                                                                                                                                                         |\n| -d, --downloader [ me \\| aget_py \\| aget_rs \\| aria2 ] | 指定下载应用<br> <br> 默认为 me (AliPCS-Py 自己的下载器，支持断续下载)<br> me 使用多文件并发下载。<br> <br> 除 me 外，其他下载器，不使用多文件并发下载，使用一个文件多链接下载。<br> 如果需要下载多个小文件推荐使用 me，如果需要下载少量大文件推荐使用其他下载器。<br> <br> aget_py (https://github.com/PeterDing/aget) 默认安装<br> aget_rs (下载 https://github.com/PeterDing/aget-rs/releases)<br> aria2 (下载 https://github.com/aria2/aria2/releases)<br> |\n| --encrypt-password, --ep TEXT                          | 加密密码，默认使用用户设置的                                                                                                                                                                                                                                                                                                                                                                                                                                   |\n\n## 播放媒体文件\n\n> 在使用了阿里云盘开放平台 api 时，如果播放他人分享的文件，文件会先保存到 `/__alipcs_py_temp__`，在从其中播放。\n> 播放结束后，会删除 `__alipcs_py_temp__` 中播放的文件。\n\n使用文件路径：\n\n```\nAliPCS-Py play [OPTIONS] [REMOTEPATHS]...\n```\n\n使用文件 ID：\n\n**注意：目前，使用 `--share-id` 或 `--file-id` 时，无法使用 `--use-local-server`**\n\n```\nAliPCS-Py play -i file_id1 -i file_id2 -i ...\n```\n\n`play` 命令默认播放带有媒体后缀的文件，如 `abc.mp4`, `abc.mp3`。如果需要播放的媒体文件没有用常规的媒体文件后缀，则需要加选项 `--ignore-ext`。\n\n### 选项\n\n| Option                        | Description                                          |\n| ----------------------------- | ---------------------------------------------------- |\n| -i, --file-id TEXT            | 文件 ID                                              |\n| --share-id, --si TEXT         | 播放这个分享 ID 下的文件                             |\n| --share-url, --su TEXT        | 播放这个分享 url 下的文件                            |\n| -p, --password TEXT           | 链接密码，如果没有不用设置                           |\n| -R, --recursive               | 递归播放                                             |\n| -f, --from-index INTEGER      | 从所有目录中的第几个文件开始播放，默认为 0（第一个） |\n| -I, --include TEXT            | 筛选包含这个字符串的文件                             |\n| --include-regex, --IR TEXT    | 筛选包含这个正则表达式的文件                         |\n| -E, --exclude TEXT            | 筛选 不 包含这个字符串的文件                         |\n| --exclude-regex, --ER TEXT    | 筛选 不 包含这个正则表达式的文件                     |\n| --player-params, --PP TEXT    | 第三方播放器参数                                     |\n| -q, --quiet                   | 取消第三方播放器输出                                 |\n| --shuffle, --sf               | 随机播放                                             |\n| --ignore-ext, --IE            | 不用文件名后缀名来判断媒体文件                       |\n| --out-cmd, --OC               | 输出第三方播放器命令                                 |\n| --pl, --player [mpv]          | 指定第三方播放器<br><br>默认为 mpv (https://mpv.io)  |\n| -s, --use-local-server        | 使用本地服务器播放。                                 |\n| --encrypt-password, --ep TEXT | 加密密码，默认使用用户设置的                         |\n\n## 上传文件\n\n上传一些本地文件或目录到网盘目录。\n\nAliPCS-Py 首先会尝试秒传。如果秒传失败，会使用分片上传上传文件。\n\n上传过程中，按 “p” 可以暂停或继续上传。\n\n```\nAliPCS-Py upload [OPTIONS] [LOCALPATHS]... REMOTEDIR\n```\n\n指定上传方式：\n\n`--upload-type Many`: 同时上传多个文件。\n\n适合大多数文件长度小于 100M 以下的情况。\n\n```\nAliPCS-Py upload --upload-type Many [OPTIONS] [LOCALPATHS]... REMOTEDIR\n```\n\n<del>`--upload-type One`: 一次只上传一个文件，但同时上传文件的多个分片。</del>\n\n<del>适合大多数文件长度大于 1G 以上的情况。</del>\n\n**阿里网盘不支持单文件并发上传。`upload --upload-type One` 失效**\n\n```\nAliPCS-Py upload --upload-type One [OPTIONS] [LOCALPATHS]... REMOTEDIR\n```\n\n指定同时上传连接数量:\n\n`--max-workers` 默认为 CPU 核数。\n\n```\nAliPCS-Py upload --max-workers 4 [OPTIONS] [LOCALPATHS]... REMOTEDIR\n```\n\n如果上传本地目录 `localdir` 到远端目录 `remotedir`，AliPCS-Py 是将 `localdir` 这个目录上传到远端目录 `remotedir` 下。\n\n比如，`localdir` 下有 2 个文件 `a`，`b` 和一个下级目录 `sub/`，如果运行 `AliPCS-Py upload localdir remotedir`，结果是远端目录 `remotedir` 下增加了 1 个下级目录和它的所有文件 `localdir/a`，`localdir/b` 和一个下级目录 `localdir/sub/`。\n\n如果要将 `localdir` 下的所有文件（包括下级目录）上传到远端目录 `remotedir`，用 `AliPCS-Py upload localdir/* remotedir`\n\n### 选项\n\n| Option                                                     | Description                             |\n| ---------------------------------------------------------- | --------------------------------------- |\n| -t, --upload-type [One \\| Many]                            | 上传方式，Many (默认): 同时上传多个文件 |\n| --encrypt-password, --ep TEXT                              | 加密密码，默认使用用户设置的            |\n| -e, --encrypt-type [No \\| Simple \\| ChaCha20 \\| AES256CBC] | 文件加密方法，默认为 No 不加密          |\n| -w, --max-workers INTEGER                                  | 同时上传文件连接数量，默认为 CPU 核数   |\n| --no-ignore-existing, --NI                                 | 上传已经存在的文件                      |\n| --no-show-progress, --NP                                   | 不显示上传进度                          |\n\n## 同步本地目录到远端\n\n同步本地目录到远端。\n\n如果本地文件最后修改时间或大小和远端不同则上传文件。对于本地不存在的文件但远端存在则删除远端文件。\n\n```\nAliPCS-Py sync [OPTIONS] LOCALDIR REMOTEDIR\n```\n\n### 选项\n\n| Option                                                     | Description                    |\n| ---------------------------------------------------------- | ------------------------------ |\n| --encrypt-password, --ep TEXT                              | 加密密码，默认使用用户设置的   |\n| -e, --encrypt-type [No \\| Simple \\| ChaCha20 \\| AES256CBC] | 文件加密方法，默认为 No 不加密 |\n| -w, --max-workers INTEGER                                  | 同时上传文件数                 |\n| --no-show-progress, --NP                                   | 不显示上传进度                 |\n\n## 关于秒传连接\n\n**阿里云盘已经不能实现用秒传连接上传。**\n\n## 分享文件\n\n```\nAliPCS-Py share [OPTIONS] [REMOTEPATHS]...\n```\n\n### 选项\n\n| Option                      | Description                      |\n| --------------------------- | -------------------------------- |\n| -p, --password TEXT         | 设置秘密，4 个字符。默认没有秘密 |\n| --period-time, --pt INTEGER | 设置分享有效期，单位为天         |\n\n## 列出分享链接\n\n```\nAliPCS-Py shared\n```\n\n### 选项\n\n| Option         | Description                                  |\n| -------------- | -------------------------------------------- |\n| -A, --show-all | 显示所有分享的链接，默认只显示有效的分享链接 |\n\n## 取消分享链接\n\n```\nAliPCS-Py cancelshared [OPTIONS] [SHARE_IDS]...\n```\n\n## 列出其他用户分享链接中的文件\n\n给 `ls` 命令加参数，我们可以列出其他用户分享链接中的文件。\n\n**注意: 这里 `remotepaths` 只能是绝对路径。**\n\n如果有分享密码，加上 `--password PASSWORD`。\n\n- 使用分享连接\n\n  ```\n  AliPCS-Py ls --share-url SHARE_URL\n\n  # e.g.\n\n  # 列出这个分享url的根目录\n  AliPCS-Py ls --share-url https://www.aliyundrive.com/s/DaNNsdvi9bi\n\n  # 递归列出这个分享url的文件\n  AliPCS-Py ls --share-url https://www.aliyundrive.com/s/DaNNsdvi9bi -R\n\n  # 列出这个分享url的指定目录\n  AliPCS-Py ls --share-url https://www.aliyundrive.com/s/DaNNsdvi9bi/folder/613397974b634251e0554d7cb56df3e46d473239\n  ```\n\n- 使用分享 ID\n\n  分享 ID 是 https://www.aliyundrive.com/s/DaNNsdvi9bi 中的 `DaNNsdvi9bi`\n\n  ```\n  AliPCS-Py ls /path --share-id SHARE_ID\n\n  # e.g.\n\n  # 列出这个分享ID的根目录\n  AliPCS-Py ls / --share-id DaNNsdvi9bi\n\n  # 列出这个分享ID下，文件名为 "613397974b634251e0554d7cb56df3e46d473239" 的文件\n  AliPCS-Py ls --file-id 613397974b634251e0554d7cb56df3e46d473239 --share-id DaNNsdvi9bi\n  ```\n\n### 选项\n\n见 `ls` 命令。\n\n## 下载他人分享的文件\n\n使用 `--share-id` 或 `--share-url` 选项，`download` 命令可以直接下载他人分享的文件。\n\n**注意: 这里必须指定下载绝对路径或 file id**\n\n```\nAliPCS-Py download /path --share-url SHARED_URL\n\n# or\n\nAliPCS-Py download /path --share-id SHARED_URL\n\n# or\n\nAliPCS-Py download --file-id FILE_ID --share-id SHARED_URL\n```\n\n- 使用分享连接\n\n  ```\n  # 下载这个分享url下的根目录中的文件\n  AliPCS-Py download / --share-url https://www.aliyundrive.com/s/DaNNsdvi9bi\n\n  # 递归下载这个分享url下的所有的文件\n  AliPCS-Py download / --share-url https://www.aliyundrive.com/s/DaNNsdvi9bi -R\n\n  # 下载这个分享url的指定目录\n  AliPCS-Py download --share-url https://www.aliyundrive.com/s/DaNNsdvi9bi/folder/613397974b634251e0554d7cb56df3e46d473239\n  ```\n\n- 使用分享 ID\n\n  ```\n  # 下载这个分享url下的根目录中的文件\n  AliPCS-Py download / --share-id DaNNsdvi9bi\n\n  # 递归下载这个分享url下的所有的文件\n  AliPCS-Py download / --share-id DaNNsdvi9bi -R\n\n  # 下载这个分享ID下，文件名为 "613397974b634251e0554d7cb56df3e46d473239" 的文件\n  AliPCS-Py download --file-id 613397974b634251e0554d7cb56df3e46d473239 --share-id DaNNsdvi9bi\n  ```\n\n### 选项\n\n见 `download` 命令。\n\n## 播放他人分享的文件\n\n使用 `--share-id` 或 `--share-url` 选项，`play` 命令可以直接播放他人分享的媒体文件。\n\n**注意: 这里必须指定下载绝对路径或 file id**\n\n```\nAliPCS-Py play /path --share-url SHARED_URL\n\n# or\n\nAliPCS-Py play /path --share-id SHARED_URL\n\n# or\n\nAliPCS-Py play --file-id FILE_ID --share-id SHARED_URL\n```\n\n- 使用分享连接\n\n  ```\n  # 播放这个分享url下的根目录中的文件\n  AliPCS-Py play / --share-url https://www.aliyundrive.com/s/DaNNsdvi9bi\n\n  # 递归播放这个分享url下的所有的文件\n  AliPCS-Py play / --share-url https://www.aliyundrive.com/s/DaNNsdvi9bi -R\n\n  # 播放这个分享url的指定目录\n  AliPCS-Py play --share-url https://www.aliyundrive.com/s/DaNNsdvi9bi/folder/613397974b634251e0554d7cb56df3e46d473239\n  ```\n\n- 使用分享 ID\n\n  ```\n  # 播放这个分享url下的根目录中的文件\n  AliPCS-Py play --share-id DaNNsdvi9bi\n\n  # 递归播放这个分享url下的所有的文件\n  AliPCS-Py play --share-id DaNNsdvi9bi -R\n\n  # 播放这个分享ID下，文件名为 "613397974b634251e0554d7cb56df3e46d473239" 的文件\n  AliPCS-Py play --file-id 613397974b634251e0554d7cb56df3e46d473239 --share-id DaNNsdvi9bi\n  ```\n\n### 选项\n\n见 `play` 命令。\n\n## 保存其他用户分享的链接\n\n保存其他用户分享的链接到远端目录。\n\n```\nAliPCS-Py save [OPTIONS] SHARE_URL_OR_ID REMOTEDIR\n```\n\n指定 `--file-id` 可以保存指定的文件。\n\n### 选项\n\n| Option              | Description                    |\n| ------------------- | ------------------------------ |\n| -i, --file-id TEXT  | 文件 ID                        |\n| -p, --password TEXT | 分享链接密码，如果没有不用设置 |\n\n## 本地保存分享连接\n\n可以将他人分享了连接保存至本地，而不需要保存在网盘。这只作为一个记录。在需要是提供查看搜索功能。\n\n使用这个功能，需要使用者在本地配置文件(`~/.alipcs-py/config.toml`)中配置:\n\n```toml\n[share]\nstore = true\n```\n\n这个功能开启后，所有与他人分享连接的操作，都会将连接和其中访问过的文件信息保存在本地文件 `~/.alipcs-py/shared-store.sqlite3` 中。\n\n使用者可以用下面的命令来查看或搜索保存的分享连接。\n\n## 保存分享连接至本地\n\n```\nAliPCS-Py storesharedlinks [OPTIONS] [SHARE_URLS_OR_IDS]...\n```\n\n**注意**: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true\n\n| Option              | Description                    |\n| ------------------- | ------------------------------ |\n| -p, --password TEXT | 分享链接密码，如果没有不用设置 |\n\n## 显示本地保存的分享连接\n\n```\nAliPCS-Py listsharedlinks\n```\n\n**注意**: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true\n\n## 显示本地保存的分享文件\n\n```\nAliPCS-Py listsharedfiles [OPTIONS]\n```\n\n**注意**: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true\n\n| Option                | Description                |\n| --------------------- | -------------------------- |\n| --share-id, --si TEXT | 指定显示 share id 下的文件 |\n\n## 查找本地保存的分享连接\n\n```\nAliPCS-Py findsharedlinks [KEYWORDS]...\n```\n\n**注意**: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true\n\n## 查找本地保存的分享文件\n\n```\nAliPCS-Py findsharedfiles [KEYWORDS]...\n```\n\n**注意**: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true\n\n| Option                | Description       |\n| --------------------- | ----------------- |\n| --share-id, --si TEXT | 要搜索的 share id |\n| -v, --verbose         | 显示细节          |\n\n## 查找本地保存的分享连接和文件\n\n```\nAliPCS-Py findshared [OPTIONS] [KEYWORDS]...\n```\n\n**注意**: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true\n\n| Option        | Description |\n| ------------- | ----------- |\n| -v, --verbose | 显示细节    |\n\n## 删除本地保存的分享连接或文件\n\n```\nAliPCS-Py deletestoredshared [OPTIONS] [SHARE_IDS]...\n```\n\n**注意**: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true\n\n| Option             | Description                                                |\n| ------------------ | ---------------------------------------------------------- |\n| -k, --keyword TEXT | 要删除文件名的关键字，如果为空则删除 share_id 下的所有文件 |\n\n## 清理本地保存的无效分享连接\n\n```\nAliPCS-Py cleanstore\n```\n\n**注意**: 使用这个命令必须将配置文件 ~/.alipcs-py/config.toml 中的 [share] store 设为 true\n\n## 开启 HTTP 服务\n\n在远端 `ROOT_DIR` 目录下开启 HTTP 服务。\n\n`ROOT_DIR` 默认为 `/`\n\n```\nAliPCS-Py AliPCS-Py server [OPTIONS] [ROOT_DIR]\n```\n\n如果需要设置认证，使用下面的选项设置用户名和密钥：\n\n```\nAliPCS-Py AliPCS-Py server [ROOT_DIR] --username \'foo\' --password \'bar\'\n```\n\n也可以指定服务路径：\n\n```\nAliPCS-Py AliPCS-Py server [ROOT_DIR] --path \'/my/pan\'\n\n# 访问 http://localhost:8000/my/pan/\n```\n\n### 选项\n\n| Option                        | Description                  |\n| ----------------------------- | ---------------------------- |\n| --path TEXT                   | 服务路径，默认为 “/”         |\n| -h, --host TEXT               | 监听 host                    |\n| -p, --port INTEGER            | 监听 port                    |\n| -w, --workers INTEGER         | 进程数                       |\n| --encrypt-password, --ep TEXT | 加密密码，默认使用用户设置的 |\n| --username TEXT               | HTTP Basic Auth 用户名       |\n| --password TEXT               | HTTP Basic Auth 密钥         |\n',
    'author': 'PeterDing',
    'author_email': 'dfhayst@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/PeterDing/AliPCS-Py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from typing import Optional, Callable, Any
from functools import wraps
from threading import Semaphore


def sure_release(semaphore: Semaphore, func, *args, **kwargs):
    try:
        return func(*args, **kwargs)
    finally:
        semaphore.release()


def retry(times: int, except_callback: Optional[Callable[..., Any]] = None):
    """Retry times when func fails"""

    def wrap(func):
        @wraps(func)
        def retry_it(*args, **kwargs):
            nonlocal times
            if times < 0:  # forever
                times = 1 << 32

            for i in range(1, times + 1):
                try:
                    r = func(*args, **kwargs)
                    return r
                except Exception as err:
                    if except_callback is not None:
                        except_callback(err, i)

                    if i == times:
                        raise err

        return retry_it

    return wrap



import random
import copy

cdef crypt(unsigned char *data, unsigned char *byte_map, int len_):
    cdef unsigned int c

    for i in range(len_):
        c = data[i]
        data[i] = byte_map[c]


class SimpleCryptography:
    """Simple Cryptography

    This crypto algorithm uses a random uint8 map to transfer an uint8 to another uint8.
    So, the decryption process does not depend on previous decrypted data.

    The algorithm is vulnerable, so NO using to encrypt important data.
    """

    def __init__(self, key):
        rg = random.Random()
        rg.seed(key, version=2)

        _byte_map = list(range(1 << 8))
        rg.shuffle(_byte_map)

        # encrypt_byte_map[ori_char] -> encrypted_char
        self._encrypt_byte_map = bytes(bytearray(_byte_map))

        # decrypt_byte_map[encrypted_char] -> ori_char
        self._decrypt_byte_map = bytes(
            bytearray([c for _, c in sorted(zip(_byte_map, range(1 << 8)))])
        )

        self._key = key

    def encrypt(self, data):
        data = bytes(bytearray(data))  # copy
        crypt(data, self._encrypt_byte_map, len(data))
        return data

    def decrypt(self, data):
        data = bytes(bytearray(data))  # copy
        crypt(data, self._decrypt_byte_map, len(data))
        return data

    def reset(self):
        pass




from typing import Optional, List

from alipcs_py.alipcs import AliPCSApi, AliPCSError
from alipcs_py.alipcs.inner import PcsFile, FromTo
from alipcs_py.common.path import (
    posix_path_dirname,
    split_posix_path,
    posix_path_basename,
    join_path,
)
from alipcs_py.commands.display import display_files, display_from_to


def makedir(api: AliPCSApi, *remotedirs: str, show: bool = False) -> List[PcsFile]:
    pcs_files = []
    for d in remotedirs:
        pf = api.makedir_path(d)
        pcs_files.append(pf)

    if show:
        display_files(pcs_files, "", show_file_id=True, show_absolute_path=True)

    return pcs_files


def move(api: AliPCSApi, *remotepaths: str, show: bool = False):
    assert len(remotepaths) > 1

    dest_remotepath = remotepaths[-1]
    pcs_files = api.paths(*remotepaths)

    # Make sure destination directory exists
    if not pcs_files[-1]:
        dest_pcs_file = api.makedir_path(dest_remotepath)
        pcs_files[-1] = dest_pcs_file

    file_ids = [pf.file_id for pf in pcs_files if pf]
    oks = api.move(*file_ids)

    from_to_list = []
    j = 0
    for i, pf in enumerate(pcs_files[:-1]):
        if not pf:
            continue

        if oks[j]:
            from_to = FromTo(
                from_=remotepaths[i],
                to_=join_path(dest_remotepath, posix_path_basename(remotepaths[i])),
            )
            from_to_list.append(from_to)
        j += 1

    if show:
        display_from_to(*from_to_list)


def rename(api: AliPCSApi, remotepath: str, new_name: str, show: bool = False):
    pcs_file = api.path(remotepath)

    if not pcs_file:
        return

    api.rename(pcs_file.file_id, new_name)
    pcs_file.path = join_path(*split_posix_path(pcs_file.path)[:-1], new_name)

    if show:
        display_files([pcs_file], "", show_file_id=True, show_absolute_path=True)


def copy(api: AliPCSApi, *remotepaths: str, show: bool = False):
    assert len(remotepaths) > 1

    dest_remotepath = remotepaths[-1]
    pcs_files = api.paths(*remotepaths)

    # Make sure destination directory exists
    if not pcs_files[-1]:
        dest_pcs_file = api.makedir_path(dest_remotepath)
        pcs_files[-1] = dest_pcs_file

    file_ids = [pf.file_id for pf in pcs_files if pf]

    new_pfs = api.copy(*file_ids)

    if show:
        display_files(new_pfs, "", show_file_id=True, show_absolute_path=True)


def remove(api: AliPCSApi, *remotepaths: str):
    pcs_files = api.paths(*remotepaths)
    file_ids = [pf.file_id for pf in pcs_files if pf]
    api.remove(*file_ids)



from typing import Optional, Tuple, List, Any

from peewee import SQL

from alipcs_py.alipcs.api import AliPCSApiMix
from alipcs_py.alipcs.inner import PcsSharedLinkInfo, PcsFile
from alipcs_py.storage.tables import PcsSharedLinkInfoTable, PcsFileTable


class SharedStore:
    def __init__(self) -> None:
        pass

    def get_shared_link_info(self, share_id: str) -> Optional[PcsSharedLinkInfo]:
        ins = PcsSharedLinkInfoTable.get_or_none(share_id=share_id)
        if ins:
            return ins.to_pcs()
        else:
            return None

    def add_shared_link_info(self, pcs_shared_link_info: PcsSharedLinkInfo) -> Any:
        pcs_shared_link_info_ins, _ = PcsSharedLinkInfoTable.get_or_create(
            **{k: getattr(pcs_shared_link_info, k) for k in pcs_shared_link_info.__dataclass_fields__}
        )
        return pcs_shared_link_info_ins

    def add_shared_file(self, share_id: str, pcs_file: PcsFile) -> Any:
        if PcsFileTable.get_or_none(file_id=pcs_file.file_id):
            # The file exists
            return

        pcs_shared_link_info_ins = PcsSharedLinkInfoTable.get_or_none(share_id=share_id)

        pcs_file_ins, _ = PcsFileTable.get_or_create(
            **{k: getattr(pcs_file, k) for k in pcs_file.__dataclass_fields__},
            shared_link_info_id=pcs_shared_link_info_ins.id,
        )
        return pcs_file_ins

    def delete_shared_links(self, *share_ids: str) -> None:
        PcsSharedLinkInfoTable.delete().where(PcsSharedLinkInfoTable.share_id.in_(share_ids)).execute()

    def delete_shared_files(self, *file_ids: str) -> None:
        PcsFileTable.delete().where(PcsFileTable.file_id.in_(file_ids)).execute()

    def search_shared_links(
        self, *keywords: str, fields: List[str] = ["share_name", "display_name"]
    ) -> List[PcsSharedLinkInfo]:
        sql = " OR ".join([f"`{f}` like ?" for f in fields * len(keywords)])
        query = PcsSharedLinkInfoTable.select().where(SQL(sql, [f"%{keyword}%" for keyword in keywords] * len(fields)))
        return [item.to_pcs() for item in query]

    def search_shared_files(
        self,
        *keywords: str,
        fields: List[str] = ["name", "path"],
        share_ids: List[str] = [],
    ) -> List[Tuple[PcsFile, PcsSharedLinkInfo]]:
        sql = " OR ".join([f"`{f}` like ?" for f in fields * len(keywords)])
        query = PcsFileTable.select().join(
            PcsSharedLinkInfoTable,
            on=(PcsFileTable.shared_link_info_id == PcsSharedLinkInfoTable.id),
        )
        if share_ids:
            query = query.where(
                PcsSharedLinkInfoTable.share_id.in_(share_ids)
                & SQL(sql, [f"%{keyword}%" for keyword in keywords] * len(fields))
            )
        else:
            query = query.where(SQL(sql, [f"%{keyword}%" for keyword in keywords] * len(fields)))
        return [(item.to_pcs(), item.shared_link_info_id.to_pcs()) for item in query]

    def list_shared_links(
        self,
        by_id: bool = False,
        by_name: bool = False,
        limit: int = 100,
        offset: int = 0,
    ) -> List[PcsSharedLinkInfo]:
        query = PcsSharedLinkInfoTable.select()
        if by_name:
            query = query.order_by(PcsSharedLinkInfoTable.share_name)
        else:
            # Default by id
            query = query.order_by(PcsSharedLinkInfoTable.id)

        query = query.limit(limit).offset(offset)

        return [item.to_pcs() for item in query]

    def list_shared_files(
        self,
        share_ids: List[str] = [],
        by_id: bool = False,
        by_name: bool = False,
        by_path: bool = False,
        limit: int = 100,
        offset: int = 0,
    ) -> List[Tuple[PcsFile, PcsSharedLinkInfo]]:
        query = PcsFileTable.select().join(
            PcsSharedLinkInfoTable,
            on=(PcsFileTable.shared_link_info_id == PcsSharedLinkInfoTable.id),
        )
        if share_ids:
            query = query.where(PcsSharedLinkInfoTable.share_id.in_(share_ids))
        if by_name:
            query = query.order_by(PcsFileTable.name)
        elif by_path:
            query = query.order_by(PcsFileTable.path)
        else:
            # Default by id
            query = query.order_by(PcsFileTable.id)

        # If limit is 0, selecting all items
        if limit > 0:
            query = query.limit(limit).offset(offset)

        return [(item.to_pcs(), item.shared_link_info_id.to_pcs()) for item in query]


class AliPCSApiMixWithSharedStore(AliPCSApiMix):
    """AliPCS API Mix with SharedStore

    Hooking the `AliPCSApiMix.list` to store the shared file infos
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._sharedstore: Optional[SharedStore] = None

    @property
    def sharedstore(self) -> Optional[SharedStore]:
        return self._sharedstore

    def get_share_token(self, share_id: str, share_password: str = "") -> str:
        token = super().get_share_token(share_id, share_password=share_password)
        if not self._sharedstore:
            return token

        if token:
            shared_link_info = self._sharedstore.get_shared_link_info(share_id)
            if not shared_link_info:
                shared_link_info = self.shared_info(share_id)
                self._sharedstore.add_shared_link_info(shared_link_info)

        return token

    def meta(self, *args, **kwargs):
        share_id = kwargs.get("share_id")

        pcs_files = super().meta(*args, **kwargs)
        if not self._sharedstore:
            return pcs_files

        if share_id:
            for pcs_file in pcs_files:
                self._sharedstore.add_shared_file(share_id, pcs_file)

        return pcs_files

    def list(self, *args, **kwargs):
        share_id = kwargs.get("share_id")

        pcs_files, next_marker = super().list(*args, **kwargs)
        if not self._sharedstore:
            return pcs_files, next_marker

        if share_id:
            for pcs_file in pcs_files:
                self._sharedstore.add_shared_file(share_id, pcs_file)

        return pcs_files, next_marker



from pathlib import Path

# Local data directory
DATA_DIR = Path("~").expanduser() / ".alipcs-py"

# Account data path
ACCOUNT_DATA_PATH = DATA_DIR / "accounts.pk"

# Rapiduploadinfo database path
RAPIDUPLOADINFO_PATH = DATA_DIR / "rapiduploadinfo.sqlite3"

# Shared store database path
SHARED_STORE_PATH = DATA_DIR / "shared-store.sqlite3"

# Logging path
LOG_PATH = DATA_DIR / "running.log"

# Logging level
LOG_LEVEL = "CRITICAL"

# Configuration path
CONFIG_PATH = DATA_DIR / "config.toml"



from typing import Dict, Any
from collections import UserDict
from functools import wraps, _make_key
import time


class TimeoutCache(UserDict):
    def __init__(self, timeout: int):
        super().__init__()
        self._timeout = timeout
        self._last_used: Dict[Any, float] = {}

    def __getitem__(self, key):
        val = super().__getitem__(key)

        now = time.time()
        if now - self._last_used[key] > self._timeout:
            self.clear_timeout()
            raise KeyError(key)

        return val

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        now = time.time()
        self._last_used[key] = now

    def clear_timeout(self):
        now = time.time()
        for key in list(self.keys()):
            if now - self._last_used[key] > self._timeout:
                self.__delitem__(key)
                del self._last_used[key]


def timeout_cache(timeout: int):
    def cached(func):
        _cache = TimeoutCache(timeout)

        @wraps(func)
        def wrap(*args, **kwargs):
            key = _make_key(args, kwargs, False)
            val = _cache.get(key)
            if val:
                return val
            val = func(*args, **kwargs)
            _cache[key] = val
            return val

        return wrap

    return cached



from typing import Optional, List, Tuple, Dict, Union, Pattern, Any
import time


from alipcs_py.alipcs import (
    PcsFile,
    FromTo,
    PcsSharedLink,
    PcsSharedLinkInfo,
    PcsUser,
    PcsRapidUploadInfo,
)
from alipcs_py.common.path import join_path
from alipcs_py.commands.sifter import Sifter
from alipcs_py.utils import format_date, format_time, human_size

_print = print

from rich.console import Console
from rich.table import Table
from rich.box import SIMPLE, MINIMAL
from rich.text import Text
from rich.highlighter import Highlighter as RichHighlighter
from rich.panel import Panel
from rich.style import Style
from rich import print


class Highlighter(RichHighlighter):
    def __init__(self, patterns: List[Union[Pattern, str]], style: Union[str, Style]):
        super().__init__()
        self.patterns = patterns
        self.style = style

    def highlight(self, text: Text):
        for pat in self.patterns:
            if isinstance(pat, Pattern):
                for m in pat.finditer(text.plain):
                    text.stylize(style=self.style, start=m.start(), end=m.end())
            else:
                pat_len = len(pat)
                start = 0
                while True:
                    idx = text.plain.find(pat, start)
                    if idx == -1:
                        break
                    text.stylize(style=self.style, start=idx, end=idx + pat_len)
                    start = idx + 1


def display_files(
    pcs_files: List[PcsFile],
    remotepath: str = "",
    sifters: List[Sifter] = [],
    highlight: bool = False,
    show_size: bool = False,
    show_date: bool = False,
    show_file_id: bool = False,
    show_hash: bool = False,
    show_absolute_path: bool = False,
    show_dl_link: bool = False,
    csv: bool = False,
):
    if not pcs_files:
        return

    table = Table(box=SIMPLE, padding=0, show_edge=False)
    table.add_column()
    headers = []  # for csv
    headers.append("\t")
    if show_size:
        header = "Size"
        table.add_column(header, justify="right", overflow="fold")
        headers.append(header)
    if show_date:
        header = "Updated Time"
        table.add_column(header, justify="center", overflow="fold")
        headers.append(header)
    if show_file_id:
        header = "File ID"
        table.add_column(header, justify="left", overflow="fold")
        headers.append(header)
    if show_hash:
        header = "SHA1"
        table.add_column(header, justify="left", overflow="fold")
        headers.append(header)
    header = "Path"
    table.add_column(header, justify="left", overflow="fold")
    headers.append(header)
    if show_dl_link:
        header = "Download Link"
        table.add_column(header, justify="left", overflow="fold")
        headers.append(header)

    rows = []  # for csv

    max_size_str_len = max([len(str(pcs_file.size)) for pcs_file in pcs_files])
    for pcs_file in pcs_files:
        row: List[Union[str, Text]] = []

        if csv:
            row.append("-")
        else:
            tp = Text("-", style="bold red")
            row.append(tp)

        if show_size:
            size = human_size(pcs_file.size) if pcs_file.size else ""
            if csv:
                row.append(f"{size} {pcs_file.size}")
            else:
                row.append(f"{size} {pcs_file.size or 0: >{max_size_str_len}}")
        if show_date:
            date = format_date(pcs_file.updated_at) if pcs_file.updated_at else ""
            row.append(date)
        if show_file_id:
            file_id = pcs_file.file_id or ""
            row.append(file_id)
        if show_hash:
            rapid_upload_info = pcs_file.rapid_upload_info
            content_hash = rapid_upload_info and rapid_upload_info.content_hash or ""
            row.append(content_hash)

        if show_absolute_path:
            path = pcs_file.path
            if not path:
                path = join_path(remotepath, pcs_file.name)
        else:
            path = pcs_file.name
        background = Text()
        if pcs_file.is_dir:
            if csv:
                row[0] = "d"
            else:
                tp._text = ["d"]
                background.style = "blue"

        if highlight and sifters:
            pats: List[Union[Pattern, str]] = list(
                filter(None, [sifter.pattern() for sifter in sifters if sifter.include()])
            )
            highlighter = Highlighter(pats, "yellow")
            _path = highlighter(path)
        else:
            _path = Text(path)

        if csv:
            row.append(path)
        else:
            row.append(background + _path)

        if show_dl_link:
            row.append(pcs_file.download_url or "")

        if csv:
            rows.append(row)
        else:
            table.add_row(*row)

    if csv:
        _print(remotepath)
        _print("\t".join(headers))
        for row in rows:
            _print("\t".join(row))  # type: ignore
    else:
        console = Console()
        if remotepath:
            title = Text(remotepath, style="italic green")
            console.print(title)
        console.print(table)


def display_from_to(*from_to_list: FromTo):
    if not from_to_list:
        return

    table = Table(box=SIMPLE, padding=0, show_edge=False)
    table.add_column("From", justify="left", overflow="fold")
    table.add_column("To", justify="left", overflow="fold")

    for from_to in from_to_list:
        table.add_row(from_to.from_, from_to.to_)

    console = Console()
    console.print(table)


_SHARED_LINK_INFO_FORMAT = (
    "share id: {share_id}\n"
    "shared url: [bold]{url}[/bold]\n"
    "time left: [i yellow]{expired_time}[/i yellow]\n"
    "password: [bold red]{password}[/bold red]\n"
    "name: {name}"
)


def display_shared_links(*shared_links: PcsSharedLink):
    panels = []
    for shared_link in shared_links:
        share_id = shared_link.share_id
        url = shared_link.share_url

        if shared_link.expired:
            expired_time = "已经超期"
        elif not shared_link.expiration:
            expired_time = "永久"
        else:
            expired_time = format_time((shared_link.expiration or 0) - int(time.time()))

        password = shared_link.share_pwd or ""
        name = shared_link.share_name

        panel = Panel(
            _SHARED_LINK_INFO_FORMAT.format(
                share_id=share_id,
                url=url,
                expired_time=expired_time,
                password=password,
                name=name,
            ),
            highlight=True,
        )
        panels.append(panel)

    console = Console()
    console.print(*panels)


def display_invalid_shared_link_infos(*shared_links: PcsSharedLinkInfo):
    for shared_link in shared_links:
        print(f"[i yellow]Remove[/i yellow]: {shared_link.share_url}\t{shared_link.share_name}")


def display_shared_paths(*shared_paths):
    table = Table(box=SIMPLE, padding=0, show_edge=False)
    table.add_column()
    table.add_column("Size", justify="right")
    table.add_column("Path", justify="left", overflow="fold")

    max_size_str_len = max([len(str(shared_path.size or 0)) for shared_path in shared_paths])
    for shared_path in shared_paths:
        row: List[Union[str, Text]] = []

        # Is file
        tp = Text("-", style="bold red")
        row.append(tp)

        size = human_size(shared_path.size) if shared_path.size else ""
        row.append(f"{size} {shared_path.size or 0: >{max_size_str_len}}")

        path = shared_path.path
        background = Text()
        if shared_path.is_dir:
            tp._text = ["d"]
            background.style = "blue"

        _path = Text(path)
        row.append(background + _path)

        table.add_row(*row)

    console = Console()
    console.print(table)


def display_shared_link_infos(*shared_link_infos: PcsSharedLinkInfo):
    table = Table(box=SIMPLE, show_edge=False, highlight=True)
    table.add_column("Share URL", justify="left")
    table.add_column("Password", justify="left")
    table.add_column("Name", justify="left", overflow="fold")
    table.add_column("Expiration", justify="left", overflow="fold")

    for shared_link_info in shared_link_infos:
        table.add_row(
            f"https://www.aliyundrive.com/s/{shared_link_info.share_id}",
            shared_link_info.share_pwd or "",
            shared_link_info.share_name or shared_link_info.display_name or "",
            format_date(shared_link_info.expiration) if shared_link_info.expiration else "Never",
        )

    console = Console()
    console.print(table)


_SHARED_FILE_FORMAT = (
    "Share URL: {share_url}\n"
    "Password: [red]{password}[/red]\n"
    "Share Name: {share_name}\n"
    "Type: [yellow]{type}[/yellow]\n"
    "Size: {size}\n"
    "File ID: {file_id}\n"
    "Name: {name}\n"
    "Share URL Directory: {share_url_directory}"
)


def display_shared_files(*shared_file_and_links: Tuple[PcsFile, PcsSharedLinkInfo], verbose: bool = False):
    if verbose:
        panels = []
        for shared_file, shared_link_info in shared_file_and_links:
            if shared_file.is_dir:
                share_url = f"https://www.aliyundrive.com/s/{shared_link_info.share_id}/folder/{shared_file.file_id}"
            else:
                share_url = f"https://www.aliyundrive.com/s/{shared_link_info.share_id}"
            panel = Panel(
                _SHARED_FILE_FORMAT.format(
                    share_url=share_url,
                    password=shared_link_info.share_pwd or "",
                    share_name=shared_link_info.share_name or shared_link_info.display_name or "",
                    type=shared_file.type,
                    size=human_size(shared_file.size or 0),
                    file_id=shared_file.file_id,
                    name=shared_file.name,
                    share_url_directory=f"https://www.aliyundrive.com/s/{shared_link_info.share_id}/folder/{shared_file.parent_file_id}",
                ),
                highlight=True,
            )
            panels.append(panel)
        console = Console()
        console.print(*panels)
    else:
        table = Table(box=SIMPLE, show_edge=False, highlight=True)
        table.add_column("Share URL", justify="left")
        table.add_column("Password", justify="left")
        table.add_column("Share Name", justify="left", overflow="fold")
        table.add_column("Type", justify="left")
        table.add_column("Size", justify="right")
        table.add_column("Name", justify="left", overflow="fold")

        for shared_file, shared_link_info in shared_file_and_links:
            table.add_row(
                f"https://www.aliyundrive.com/s/{shared_link_info.share_id}",
                shared_link_info.share_pwd or "",
                shared_link_info.share_name or shared_link_info.display_name or "",
                shared_file.type,
                human_size(shared_file.size or 0),
                shared_file.name,
                #  format_date(shared_link_info.expiration)
                #  if shared_link_info.expiration
                #  else "Never",
            )

        console = Console()
        console.print(table)


def display_user_info(user_info: PcsUser):
    user_id = user_info.user_id
    default_drive_id = user_info.default_drive_id
    domain_id = user_info.domain_id
    user_name = user_info.user_name
    nick_name = user_info.nick_name
    phone = user_info.phone
    personal_space_info = user_info.personal_space_info

    web_refresh_token = user_info.web_refresh_token
    web_access_token = user_info.web_access_token
    web_token_type = user_info.web_token_type
    web_expire_time = format_date(user_info.web_expire_time or 0)

    openapi_refresh_token = user_info.openapi_refresh_token
    openapi_access_token = user_info.openapi_access_token
    openapi_token_type = user_info.openapi_token_type
    openapi_expire_time = format_date(user_info.openapi_expire_time or 0)

    client_id = user_info.client_id
    client_secret = user_info.client_secret
    client_server = user_info.client_server

    assert personal_space_info
    quota_str = human_size(personal_space_info.used_size) + "/" + human_size(personal_space_info.total_size)

    _tempt = (
        f"user id: {user_id}\n"
        f"user name: {user_name}\n"
        f"nick name: {nick_name}\n"
        f"user phone: {phone}\n"
        f"default drive id: {default_drive_id}\n"
        f"domain id: {domain_id}\n"
        f"quota: {quota_str}\n"
        "\n"
        f"web_refresh token: {web_refresh_token}\n"
        f"web_access token: {web_access_token}\n"
        f"web_token_type: {web_token_type}\n"
        f"web_expire time: {web_expire_time}\n"
        "\n"
        f"openapi_refresh token: {openapi_refresh_token}\n"
        f"openapi_access token: {openapi_access_token}\n"
        f"openapi_token_type: {openapi_token_type}\n"
        f"openapi_expire time: {openapi_expire_time}\n"
        "\n"
        f"client_id: {client_id}\n"
        f"client_secret: {client_secret}\n"
        f"client_server: {client_server}\n"
    )

    console = Console()
    console.print(_tempt, highlight=True)


def display_user_infos(*user_infos: Tuple[PcsUser, str, str], recent_user_id: Optional[int] = None):
    """
    Args:
        user_infos (*Tuple[PcsUser, pwd: str])
    """

    table = Table(box=SIMPLE, show_edge=False, highlight=True)
    table.add_column("Index", justify="left")
    table.add_column("Recent", justify="left")
    table.add_column("Account Name", justify="left", overflow="fold")
    table.add_column("User Name", justify="left", overflow="fold")
    table.add_column("Nick Name", justify="left", overflow="fold")
    table.add_column("Quota", justify="left")
    table.add_column("VIP", justify="left", overflow="fold")
    table.add_column("pwd", justify="left", overflow="fold")

    for idx, (user_info, pwd, account_name) in enumerate(user_infos, 1):
        is_recent = "[green]✔[/green]" if user_info.user_id == recent_user_id else ""

        user_name = user_info.user_name
        nick_name = user_info.nick_name
        personal_space_info = user_info.personal_space_info
        vip = ""
        if user_info.user_vip_info:
            vip = f"[b red]{user_info.user_vip_info.identity}[/b red]"

        assert personal_space_info
        quota_str = human_size(personal_space_info.used_size) + "/" + human_size(personal_space_info.total_size)

        table.add_row(str(idx), is_recent, account_name, user_name, nick_name, quota_str, vip, pwd)

    console = Console()
    console.print(table)


def display_blocked_remotepath(remotepath: str):
    print(f"[i yellow]Remote path is blocked[/i yellow]: {remotepath}")



from typing import List

from alipcs_py.alipcs import AliPCSApi
from alipcs_py.commands.display import display_files
from alipcs_py.commands.sifter import Sifter, IncludeSifter


def search(
    api: AliPCSApi,
    keyword: str,
    sifters: List[Sifter] = None,
    highlight: bool = False,
    show_size: bool = False,
    show_date: bool = False,
    show_file_id: bool = False,
    show_hash: bool = False,
    csv: bool = False,
):
    pcs_files = api.search_all(keyword)

    sifters = [*(sifters or []), IncludeSifter(keyword)]
    display_files(
        pcs_files,
        sifters=sifters,
        highlight=highlight,
        show_size=show_size,
        show_date=show_date,
        show_file_id=show_file_id,
        show_hash=show_hash,
        show_absolute_path=True,
        csv=csv,
    )



import time
from datetime import datetime, timezone


def now_timestamp() -> int:
    """Now timestamp (Second)"""

    return int(time.time())


def iso_8601_to_timestamp(date_string: str) -> int:
    """Convert ISO 8601 datetime string to timestamp

    Args:
        date_string (str): ISO 8601 format.
            e.g. "2021-06-22T07:16:03Z" or "2021-06-22T07:16:03.032Z"
    """

    if len(date_string) == 20:
        return int(datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S%z").timestamp())
    else:
        return int(datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S.%f%z").timestamp())


def timestamp_to_iso_8601(timestamp: int) -> str:
    return datetime.fromtimestamp(timestamp, timezone.utc).isoformat().replace("+00:00", ".000Z")



import typing
from typing import Tuple, List, Dict, Any
from peewee import (
    Model,
    CharField,
    IntegerField,
    Database,
    SqliteDatabase,
    TextField,
    SmallIntegerField,
    ForeignKeyField,
)
from playhouse.migrate import SchemaMigrator, SqliteMigrator, migrate

from alipcs_py.alipcs.inner import PcsFile, PcsSharedLinkInfo
from alipcs_py.common.util import json_dump_values


class Deserializer:
    @classmethod
    def pcs_item(cls) -> Any:
        raise NotImplementedError()

    def to_pcs(self) -> PcsSharedLinkInfo:
        pcs_item = self.pcs_item()
        data: Dict[str, Any] = {}
        columns = self.__class__._meta.columns  # type: ignore
        pcs_fields = pcs_item.__dataclass_fields__
        for col in columns.keys():
            if col in pcs_fields:
                val = getattr(self, col)
                if pcs_fields[col].type == typing.Optional[bool] or pcs_fields[col].type == bool:
                    if val == None:
                        data[col] = None
                    else:
                        data[col] = bool(val)
                else:
                    data[col] = val
        return pcs_item(**data)


class PcsSharedLinkInfoTable(Deserializer, Model):
    share_id = CharField(null=False, unique=True)
    share_pwd = CharField(null=True)
    share_name = CharField(null=True)
    display_name = CharField(null=True)
    file_count = IntegerField(null=True)
    file_infos = TextField(null=True)  # json
    expiration = IntegerField(null=True)
    updated_at = IntegerField(null=True)
    vip = CharField(null=True)
    avatar = CharField(null=True)
    is_following_creator = SmallIntegerField(null=True)
    creator_id = CharField(null=True)
    creator_name = CharField(null=True)
    creator_phone = CharField(null=True)

    class Meta:
        indexes = (
            # create a non-unique
            (("share_id",), False),
            (("share_name",), False),
            (("display_name",), False),
        )

    @classmethod
    def get_or_create(cls, **kwargs):
        kwargs = json_dump_values(kwargs)
        return super().get_or_create(**kwargs)

    @classmethod
    def pcs_item(cls) -> Any:
        return PcsSharedLinkInfo


class PcsFileTable(Deserializer, Model):
    file_id = CharField(null=False, unique=True)
    name = CharField(null=False)
    parent_file_id = CharField(null=False)
    type = CharField(null=False)
    is_dir = SmallIntegerField(null=False)
    is_file = SmallIntegerField(null=False)
    size = IntegerField(null=True)
    path = CharField(null=True)

    created_at = IntegerField(null=True)
    updated_at = IntegerField(null=True)

    file_extension = CharField(null=True)
    content_type = CharField(null=True)
    mime_type = CharField(null=True)
    mime_extension = CharField(null=True)
    labels = CharField(null=True)

    status = CharField(null=True)
    hidden = IntegerField(null=True)
    starred = IntegerField(null=True)
    category = CharField(null=True)
    punish_flag = IntegerField(null=True)
    encrypt_mode = CharField(null=True)

    drive_id = CharField(null=True)
    domain_id = CharField(null=True)
    upload_id = CharField(null=True)
    async_task_id = CharField(null=True)

    rapid_upload_info = CharField(null=True)
    download_url = CharField(null=True)

    shared_link_info_id = ForeignKeyField(PcsSharedLinkInfoTable)

    class Meta:
        indexes = (
            # create a non-unique
            (("file_id",), False),
            (("name",), False),
            (("is_dir",), False),
            (("is_file",), False),
            (("path",), False),
            (("file_extension",), False),
        )

    @classmethod
    def get_or_create(cls, **kwargs):
        kwargs = json_dump_values(kwargs)
        return super().get_or_create(**kwargs)

    @classmethod
    def pcs_item(cls) -> Any:
        return PcsFile


def connect_sqlite(path: str) -> Tuple[Database, SchemaMigrator]:
    db = SqliteDatabase(path)
    return db, SqliteMigrator(db)


def bind_tables(tables: List[type], db: Database):
    db.bind(tables)


def create_tables(tables: List[type], db: Database):
    db.create_tables(tables)


def get_db_field(tp: type):
    if tp == str:
        return CharField(null=True)
    elif tp == typing.Optional[str]:
        return CharField(null=True)

    elif tp == typing.List[str]:
        return TextField(null=True)

    elif tp == int:
        return IntegerField(null=True)
    elif tp == typing.Optional[int]:
        return IntegerField(null=True)

    elif tp == bool:
        return IntegerField(null=True)
    elif tp == typing.Optional[int]:
        return IntegerField(null=True)

    else:
        raise ValueError("Unsupported type: %s", tp)


def modify_table(table: Any, db: Database, migrator: SchemaMigrator):
    table_name = table._meta.name  # type: ignore
    pcs_fields = table.pcs_item().__dataclass_fields__
    columns = set([c.name for c in db.get_columns(table_name)])

    for name, field in pcs_fields.items():
        if name not in columns:
            db_field = get_db_field(field.type)
            migrate(migrator.add_column(table_name, name, db_field))
